<?php 
/**
 *-----------------------------------------------------------------------------*
 * License: Concrete5 marketplace :                                            *
 * http://www.concrete5.org/help/legal/commercial_add-on_license/              *
 *                                                                             *
 * @author (Florian Delizy)                                                    *
 *                                                                             *
 * Copyright(C) 2013-2015 Florian Delizy <florian.delizy@gmail.com>            *
 *-----------------------------------------------------------------------------*
 */
namespace Concrete\Package\GntMaxmind;

defined('C5_EXECUTE') or die('Access Denied.');

use Package;
use Page;
use Block;
use Job;
use JobSet;
use Log;
use Database;
use Events;
use BlockType;
use SinglePage;

use \Illuminate\Filesystem\Filesystem;

use Concrete\Core\Attribute\Type as AttributeType;
use Concrete\Core\Attribute\Key\Category as AttributeKeyCategory;
use Concrete\Core\Attribute\Key\CollectionKey as CollectionAttributeKey;
use Concrete\Core\Routing\Router;
use Concrete\Core\Support\Facade\Route;

use Concrete\Package\GntMaxmind\Src\MaxmindDatabase;
use Concrete\Package\GntMaxmind\Src\FilesServiceProvider;
use \Symfony\Component\HttpFoundation\Session\Session as SymfonySession;



class Controller extends Package
{

    protected $pkgHandle = 'gnt_maxmind';
    protected $appVersionRequired = "5.7.5.6";
    protected $pkgVersion = "2.0.5";

    public function getPackageDescription()
    {
        return t("IP-Geo-localisation and redirection rules with MaxMind");
    }

    public function getPackageName()
    {
        return t("MaxMind IP-GeoLoc");
    }


    public function install()
    {
        $pkg = parent::install();

		// Create the storage directory for maxmind databases
		$controller = Page::getCurrentPage()->getController();

		$path = self::getDefaultDatabasePath();
		if ($controller->post('customizePath') == 1) {
			$path = $controller->post('databasePath');
		}

		$pkg->setDatabasePath($path);

		$remove = $controller->post('removeDirectory') == 1;
		$pkg->setDeleteOnUninstall($remove);

		// install block et Co

        $r = BlockType::installBlockTypeFromPackage("gnt_maxmind", $pkg);
        if (!is_null($r)) {
            Log::warning(t("Unable to install block : %s", $r));
            throw new Exception($r);
        }


		// Create the job set and add jobs to it
		$js = JobSet::add(t('Maxmind Database Update and cleanup'), $pkg);

		$job = Job::installByPackage('gnt_maxmind_download_database', $pkg);
		$js->addJob($job);

		$job = Job::installByPackage('gnt_maxmind_delete_unused_databases', $pkg);
		$js->addJob($job);

        $this->createSinglePages($pkg);
        $this->installAttributes($pkg);

    }

    public function installAttributes($pkg = null)
    {

        if (!$pkg) {
            $pkg = $this;
        }

        $at = AttributeType::getByHandle('gnt_maxmind_redirect');
        if (!$at || !intval($at->getAttributeTypeID())) {
            $at = AttributeType::add('gnt_maxmind_redirect', t('Maxmind IP GeoLoc Redirect'), $pkg);
        }

        $akc = AttributeKeyCategory::getByHandle('collection');
        if (! $akc) {
            Log::error('gnt_maxmind:'.t("Cannot associate with not existing category 'collection'"));

            return;
        }

        if (!$at->isAssociatedWithCategory($akc)) {
            $akc->associateAttributeKeyType($at);
        }

        $key = CollectionAttributeKey::getByHandle('gnt_maxmind_redirect');
        if (!$key || !intval($key->getAttributeKeyID())) {
            $desc = array( 'akHandle' => 'gnt_maxmind_redirect', 'akName' => t('Maxmind IP GeoLoc Redirect'), 'akIsSearchable' => false );
            $key = CollectionAttributeKey::add($at, $desc);
        }
    }

    public function createSinglePages($pkg = null)
    {
        if (!$pkg) {
            $pkg = $this;
        }

        # install single page

        $pages = array(
            '/dashboard/gnt_maxmind/'    => array(
                'cName' => t('MaxMind GeoIP'), 'cDescription' => t('MaxMind GeoIP menu'), 'cIsSystemPage' => 1,
            ),
            '/dashboard/gnt_maxmind/lookup'    => array(
                'cName' => t('IP GeoLoc Lookup'), 'cDescription' => t('Perform a MaxMind GeoIP Lookup'), 'cIsSystemPage' => 1,
            ),
            '/dashboard/gnt_maxmind/databases'    => array(
                'cName' => t('Maxmind Databases'), 'cDescription' => t('Manage Maxmind Databases'), 'cIsSystemPage' => 1,
            ),
            '/dashboard/gnt_maxmind/rules'    => array(
                'cName' => t('Redirect Rules'), 'cDescription' => t('Manage Redirect Rules'), 'cIsSystemPage' => 1,
            ),
            '/dashboard/gnt_maxmind/settings'    => array(
                'cName' => t('Settings'), 'cDescription' => t('Manage Global Settings'), 'cIsSystemPage' => 1,
            ),
        );

        foreach ($pages as $path => $desc) {

			$attrs = array();

			if (isset($desc['attributes'])) {
				$attrs = $desc['attributes'];
				unset($desc['attributes']);
			}

            $page = null;

            if (!$page || $page->isError()) {
                $page = Page::getByPath($path);
            }
            if (!$page || $page->isError()) {
                $page = SinglePage::add($path, $pkg);
            }
            if (!$page || $page->isError()) {
                continue;
            }

            if (!$page) {
                continue;
            }

            $page->update($desc);
            foreach ($attrs as $key => $val) {
                $ak = CollectionAttributeKey::getByHandle($key);
                if (!$ak) {
                    continue;
                }
                $page->setAttribute($ak, $val);
            }
        }
    }

    public function uninstall()
    {
		$mustDelete = $this->mustDeleteOnUninstall();
		$path = $this->getDatabasePath();

        parent::uninstall();

		if ($mustDelete) {
			$this->_deleteDatabaseDirectory($path);
		}

        $pages = array(
            '/dashboard/system/environment/gnt_maxmind',
        );
        foreach ($pages as $path) {
            $dash = Page::getByPath($path);
            if ($dash) {
                $dash->delete();
            }
        }

        $db = Database::getActiveConnection();

        $db->Execute('DROP TABLE IF EXISTS btGntMaxmind');
        $db->Execute('DROP TABLE IF EXISTS gntMaxMindRedirectRule');
        $db->Execute('DROP TABLE IF EXISTS gntMaxMindRedirectRuleSet');
        $db->Execute('DROP TABLE IF EXISTS atGntMaxmindRedirect');
    }

	public function getDatabasePath()
	{
		if (!$this->isPackageInstalled()) {
			return $this->getDefaultDatabasePath();
		}

		$config = $this->getConfig();
		$path = $config->get('database.path');
		if (!$path) {
			$this->_setDatabasePath(self::getDefaultDatabasePath());
		}
		return $config->get('database.path');
	}

	/**
	 * @fn _setDatabasePath($path, $copyExisting=false)
	 * set the new database path, create the directory if not existing, 
	 * set the parameter and copy old content if any (see copyExisting param)
	 *
	 * @param path (string) the new path (absolute path), must be writable directory
	 * @param copyExisting (boolean) copy existing databases (from old path)
	 */
	protected function _setDatabasePath($path, $copyExisting=false)
	{
		$fs = new Filesystem();


		if (!$fs->isDirectory($path)) {
			$fs->makeDirectory($path);
		}

		if (!$fs->isDirectory($path)) {
			Log::error(t("Can not set maxmind database directory to %s (not a directory, and can not create it)", $path));
			throw new \Exception(t("%s must be a directory", $path));
		}

		if (!$fs->isWritable($path)) {
			$message = t("Maxmind IP-GeoLoc set database path to %s, but the directory is not writable, automatic upload/update will be deactivated", $path);
			Log::warning($message);
			$session = new SymfonySession();
			$session->getFlashBag()->add('errormsg', $message);
		}

		$old = $this->getConfig()->get('database.path');

		if ($old && $fs->isDirectory($old) && $copyExisting) {

			foreach ($fs->glob("$old/*.mmdb") as $file) {
				$fileName = basename($file);
				$fs->copy("$file", "$path/$fileName");
			}
		}

		$this->getConfig()->save('database.path', $path);
	}

	const NO_SIMULATION = 0;
	const SIMULATION_GET = 1;
	const SIMULATION_FIXED = 2;

	public static function simulationMethodNames()
	{
		return array(
			self::NO_SIMULATION => t('No simulation, use REMOTE_IP'),
			self::SIMULATION_GET => t('Take IP from a GET VARIABLE'),
			self::SIMULATION_FIXED => t('Fixed IP for all visitors'),
		);
	}

	public function setSimulationMethod($method)
	{
		$this->getConfig()->save('debug.simulation', intval($method));
	}

	public function simulationMethod() 
	{
		return intval($this->getConfig()->get('debug.simulation'));
	}

	public function setSimulationVariableName($name)
	{
		$this->getConfig()->save('debug.varname', $name);
	}

	public function simulationVarName()
	{
		$name = $this->getConfig()->get('debug.varname');
		if ($name == '') {
			$this->getConfig()->save('debug.varname', 'MAXMIND_IP');
			$name = $this->getConfig()->get('debug.varname');
		}
		return $name;

	}

	public function setSimulationIP($ip)
	{
		$this->getConfig()->save('debug.simulationip', $ip);
	}

	public function simulationIP()
	{
		return $this->getConfig()->get('debug.simulationip');
	}

	public function setDatabasePath($path, $deleteOld=false, $copyExisting=false)
	{
		$old = $this->getDatabasePath();
		$this->_setDatabasePath($path, $copyExisting);

		if ($deleteOld) {
			$this->_deleteDatabaseDirectory($old);
		}
	}

	protected function _deleteDatabaseDirectory($path)
	{
		// sanity check:

		$fs = new Filesystem();

		if (!$fs->isDirectory($path)) {
			Log::error(t("Can not delete directory %s: not a directory", $path));
			return;
		}

		if (!$fs->isWritable($path)) {
			Log::error(t("Can not delete database directory %s: not writable", $path));
			return;
		}

		// Now check if there is any file that is not a .mmdb file (that is anything else than a database file
		// ignoring .htaccess on purpose

		foreach ($fs->allFiles($path) as $filename) {

			$name = pathinfo($filename, PATHINFO_BASENAME);

			if ('.htaccess' == strtolower($name)) {
				continue;
			}

			$ext = strtolower($fs->extension($filename));
			if (false === array_search($ext, array('mmdb', 'gz', 'mmdb.gz'))) {
				Log::error(t("Can not delete database directory %s : %s file is in the way and is not a database", $path, $filename));
				return;
			}
		}

		Log::info(t("Deleting maxmind database directory %s", $path));
		$fs->deleteDirectory($path);
	}

	public function canWriteDatabaseStorage()
	{
		$path = $this->getDatabasePath();

		$fs = new Filesystem();
		return $fs->isDirectory($path) && $fs->isWritable($path);
	}

	public static function maxUploadFileSize()
	{
		return self::humanToBytes(ini_get('upload_max_filesize'));
	}

	public static function maxPostSize()
	{
		return self::humanToBytes(ini_get('post_max_size'));
	}

	public static function memoryLimit()
	{
		return self::humanToBytes(ini_get('memory_limit'));
	}

	public static function humanToBytes($humanSize)
	{
		$n = intval(substr($humanSize, 0, -1));
		$units = array( 'k' => 1024, 'M' => 1024*1024, 'G' => 1024*1024*1024 );
		$unit = substr($humanSize, -1);
		return $n * intval(isset($units[$unit]) ? $units[$unit] : 1);
	}

	public function mustDeleteOnUninstall()
	{
		$config = $this->getConfig();
		$ret = $config->get('database.delete_on_uninstall');
		if (null === $ret) {
			$this->setDeleteOnUninstall();
		}
		return $config->get('database.delete_on_uninstall')==1?true:false;
	}

	public function setDeleteOnUninstall($delete = true)
	{
		$this->getConfig()->save('database.delete_on_uninstall', $delete?1:0);
	}


	public static function getDefaultDatabasePath()
	{
		return DIR_FILES_UPLOADED_STANDARD . "/gnt_maxmind";
	}

    public function currentMaxMindKey()
    {
		$config = $this->getConfig();

        $current = $config->get('database.current');
        if (!$current) {
            $this->useLastestInstalled();
        }

        return $config->get('database.current');
    }

    public function setCurrentMaxMindKey($k)
    {
        $list = MaxmindDatabase::listDatabases();
        if (!isset($list[$k])) {
            return;
        }
        $this->getConfig()->save('database.current', $k);
    }

    public function useLastestInstalled()
    {
        $list = MaxmindDatabase::listDatabases();
        if (!count($list)) {
            return;
        }
        $keys = array_keys($list);
        $last = $keys[count($keys)-1];

        $this->setCurrentMaxMindKey($last);
		return $last;
    }

    public function fixCurrentKey()
    {
        $list = MaxmindDatabase::listDatabases();
        if (!count($list)) {
            return;
        }

        $k = $this->currentMaxMindKey();
        if (isset($list[$k])) {
            return $k;
        }
        return $this->useLastestInstalled();
    }

    public function isDisabled()
    {
        return $this->getConfig()->get('global.disable') == true;
    }
    public function disable($dis = true)
    {
        $this->getConfig()->save('global.disable', $dis);
    }

    public function on_start()
    {
		// On package start, start all composer packages
        $fs = new Filesystem();
        $fs->requireOnce(__DIR__ . '/vendor/autoload.php');

		Events::addListener('on_before_render', array($this, 'on_before_render'));

		// ajax refresh element:
		Route::register(Router::route(array('/ajax/redirect_rule', 'gnt_maxmind')), '\Concrete\Package\GntMaxmind\Controller\Ajax\RedirectRule::view', 'GntMaxmindRedirectRule');


		// register all services:
		$fsp = new FilesServiceProvider(\Core::make('app'));
		$fsp->register();
    }

    protected static function _getAttribute()
    {
        $page = Page::getCurrentPage();
        if (!$page || $page->isError()) {
            return;
        }

        $param = $page->getAttribute('gnt_maxmind_redirect');

        if ($param && $param->activated) {
            return $param;
        }

        while ($page && !$page->isError() && (!$param || !$param->recursive || !$param->activated)) {
            $param = $page->getAttribute('gnt_maxmind_redirect');
            $page = Page::getByID($page->getCollectionParentID());
        }
        if ($param && (!$param->recursive || !$param->activated)) {
            return;
        }

        return $param;
    }

    public static function on_before_render()
    {
        $av = self::_getAttribute();
        if (!$av) {
            return;
        }
        $rs = $av->ruleSet;
        if (!$rs) {
            return;
        }
        if ($av->debug) {
            $rs->setDebug(true);
        }
        $rs->route();
    }


	public function getCurrentIP()
	{

		$method = $this->simulationMethod();
		switch ($method) {
			case self::SIMULATION_GET:
				$varName = $this->simulationVarName();
				$ip = $_REQUEST[$varName];
				break;

			case self::SIMULATION_FIXED:
				$ip = $this->simulationIP();
				break;
		}

		if (!$ip) {
			$ip = $_SERVER['REMOTE_ADDR'];
		}

		return $ip;

	}
}
// vim: set noexpandtab ts=4 ai :

