<?php 
/**
 *-----------------------------------------------------------------------------*
 * License: Concrete5 marketplace :                                            *
 * http://www.concrete5.org/help/legal/commercial_add-on_license/              *
 *                                                                             *
 * @author (Florian Delizy)                                                    *
 *                                                                             *
 * Copyright(C) 2015 Florian Delizy <florian.delizy@gmail.com>                 *
 *-----------------------------------------------------------------------------*
 */

namespace Concrete\Package\GntMaxmind\Src\Files;

defined('C5_EXECUTE') or die('Access Denied.');

use Concrete\Core\Foundation\Object;
use Concrete\Core\Package\PackageList;
use Log;

class FileLister
{
	public function listFiles( $dirPrefix, $type = '*.php', $removeSuffix = '', $includePackages = false, $simplify = true )
	{
		# eg: listFiles( 'blocks/btHandle/dir', '*.php', '', true )

		$pkgHandle = 'gnt_maxmind';

		$paths = array();

		if ( $includePackages )
		{
			$pl = PackageList::get();
			foreach ( $pl->getPackages() as $pkg )
			{
				$handle = $pkg->getPackageHandle();
				if ( $handle == $pkgHandle ) continue;

				$path = DIR_PACKAGES . "/$handle/$dirPrefix";
				if ( is_dir( $path ) ) $paths[] = $path;
			}
		}

		# site override always override last
		$paths[] = DIR_PACKAGES . "/$pkgHandle/$dirPrefix";

		$paths[] = DIR_BASE . "/$dirPrefix";
		$paths[] = DIR_BASE . "/$handle/$dirPrefix";

		$paths[] = DIR_APPLICATION . "/$dirPrefix";
		$paths[] = DIR_APPLICATION . "/$handle/$dirPrefix";


		$files = array();

		foreach ( $paths as $path )
		{
			$fileList = glob( "$path/$type" );
			if ( false === $fileList || !is_array( $fileList ) ) continue;

			foreach ( $fileList as $file )
			{
				$info = pathinfo($file);
				$base = $info['filename'];

				if ( $removeSuffix ) $base = preg_replace ( ",$removeSuffix\$,", '', $base );
				$name = str_replace( '_', ' ', $base  );

				if ( $simplify ) $files[$name] = $file;
				else
				{
					if ( !is_array( $files[$name] ) ) $files[$name] = array();
					$files[$name][] = $file;
				}
			}
		}
		return $files;

	}

	public function selectFilesBasename( $dirPrefix, $type = '*.php', $removeSuffix = '', $includePackages = false )
	{
		$sel = array();
		$files = self::listFiles( $dirPrefix, $type, $removeSuffix, $includePackages, true );
		foreach ( $files as $name => $file )
		{
			$info = pathinfo($file);
			$base = $info['filename'];
			$name[0] = strtoupper($name[0]);
			$sel[$base] = str_replace( '_', ' ', Object::uncamelcase( $name ));
		}

		return $sel;
	}
}



// vim: set noexpandtab ts=4 :
