<?php 
/**
 *-----------------------------------------------------------------------------*
 * License: Concrete5 marketplace :                                            *
 * http://www.concrete5.org/help/legal/commercial_add-on_license/              *
 *                                                                             *
 * @author (Florian Delizy)                                                    *
 *                                                                             *
 * Copyright(C) 2015 Florian Delizy <florian.delizy@gmail.com>                 *
 *-----------------------------------------------------------------------------*
 */
namespace Concrete\Package\GntMaxmind\Src;

defined('C5_EXECUTE') or die('Access Denied.');

use Database;
use User;
use Page;
use Log;
use Core;
use View;
use URL;
use Package;
use Concrete\Core\Controller\AbstractController;

class RedirectRule
{
    const NO_REDIRECT      = 0;
    const REDIRECT_FORCED  = 1;
    const REDIRECT_SUGGEST = 2;

    public static function getRedirectTypes()
    {
        return array(
            self::NO_REDIRECT       => t('No Redirection'),
            self::REDIRECT_FORCED   => t('Forced HTTP Redirection'),
            self::REDIRECT_SUGGEST  => t('Suggest Redirect'),
        );
    }

    const TARGET_NONE      = 0;
    const TARGET_PAGE      = 1;
    const TARGET_URL       = 2;

    public static function getRedirectTargets()
    {
        return array(
            self::TARGET_NONE   => t('To nowhere (no redirection)'),
            self::TARGET_PAGE   => t('To a page'),
            self::TARGET_URL    => t('To a remote url'),
        );
    }

    const MATCH_NOMATCH     = 0;
    const MATCH_EQUAL       = 1;
    const MATCH_DIFFER      = 2;
    const MATCH_REGEX       = 3;

    public static function getMatchTypes()
    {
        return array(
            self::MATCH_NOMATCH => t('No match'),
            self::MATCH_EQUAL   => t('Equal/In'),
            self::MATCH_DIFFER  => t('Not Equal/Not In'),
            self::MATCH_REGEX   => t('Regex'),
        );
    }

    const DATUM_IP      = 0;
    const DATUM_CITY    = 1;
    const DATUM_STATE   = 2;
    const DATUM_COUNTRY = 3;

    public static function getAvailableMatchTypes($datum)
    {
        // placeholder for future specific rules
        return self::getMatchTypes();
    }

    public static function getNotifyTypes()
    {
        return array(
            'alert'   => t('alert'),
            'info'    => t('info'),
            'success' => t('success'),
            'warning' => t('warning'),
            'error'   => t('error'),
        );
    }

    public static $_layouts = null;
    public static function getNotifyLayouts()
    {
        if (!self::$_layouts) {
            $fl = \Core::make('gnt_maxmind/helpers/file_lister');
            self::$_layouts = $fl->selectFilesBasename('js/3rdparty/noty/layouts', '*.js', '', true);
        }

        return self::$_layouts;
    }

    public static $_themes = null;
    public static function getNotifyThemes()
    {
        if (!self::$_themes) {
            $fl = \Core::make('gnt_maxmind/helpers/file_lister');
            self::$_themes = $fl->selectFilesBasename('js/3rdparty/noty/themes', '*.js', '', true);
        }

        return self::$_themes;
    }

    protected $_rID              = null;
    protected $_rsID             = null;
    protected $_rank             = 0;
    protected $_countryMatchType = self::MATCH_EQUAL;
    protected $_countryCode      = null;
    protected $_stateMatchType   = self::MATCH_NOMATCH;
    protected $_stateCode        = null;
    protected $_cityMatchType    = self::MATCH_NOMATCH;
    protected $_cityName         = null;
    protected $_ipMatchType      = self::MATCH_NOMATCH;
    protected $_ip               = null;
    protected $_redirectType     = self::REDIRECT_SUGGEST;
    protected $_excludeUsers     = 1;
    protected $_excludeAdmins    = 1;
    protected $_targetType       = self::TARGET_NONE;
    protected $_targetCID        = null;
    protected $_timeout          = 0;
    protected $_link             = null;
    protected $_params           = null;
    protected $_message          = null;
    protected $_activated        = 1;
    protected $_notifyLayout     = 'top';
    protected $_notifyModal      = 0;
    protected $_notifyType       = 'alert';
    protected $_notifyTheme      = 'default';
    protected $_hideOnTimeout    = 1;
    protected $_useDefaultMessage = 1;
    protected $_formID            = null;

    public function ruleID()
    {
        return $this->_rID;
    }

    public function formID()
    {
        if ($this->_rID) {
            return $this->_rID;
        }
        if ($this->_formID) {
            return $this->_formID;
        }
        $this->_formID = uniqid();

        return $this->_formID;
    }

    public function setFormID($formID)
    {
        $this->_formID = $formID;
    }

    protected $_ruleSet            = null;
    public function setRuleSet($rs)
    {
        if (is_object($rs)) {
            $ruleSet = $rs;
            $rsID = $rs->getRuleSetID();
        } else {
            $rsID = $rs;
            $ruleSet = RedirectRuleSet::getByID($rs);
        }
        $this->_rsID = $rsID;
        $this->_ruleSet = $ruleSet;
    }

    public function getRuleSet()
    {
        if (!$this->_ruleSet && $this->_rsID) {
            $this->_ruleSet = RedirectRuleSet::getByID($this->_rsID);
        }

        return $this->_ruleSet;
    }

    public function debug()
    {
        $rs = $this->getRuleSet();
        if ($rs) {
            return $rs->debug();
        }

        return false;
    }

    public function debugLog($msg)
    {
        $rs = $this->getRuleSet();
        if ($rs) {
            $rs->debugLog($this->_rank, $msg);
        }
    }

    public function __construct($rsID = null)
    {
        if ($rsID) {
            $this->setRuleSet($rsID);
        }
    }

    public function notifyModal()
    {
        return $this->_notifyModal ? true : false;
    }
    public function notifyLayout()
    {
        return $this->_notifyLayout;
    }
    public function notifyTheme()
    {
        return $this->_notifyTheme;
    }
    public function notifyType()
    {
        return $this->_notifyType;
    }
    public function timeout()
    {
        return $this->_timeout;
    }
    public function hideOnTimeout()
    {
        return $this->_hideOnTimeout;
    }
    public function rsID()
    {
        return $this->_rsID;
    }

    protected $_cookedMessage = null;
    public function message()
    {
        if (!$this->_cookedMessage) {
            $message = $this->_message;
            if ($this->_useDefaultMessage) {
                $message = $this->getRuleSet()->message();
            }

            $message = MaxmindDatabase::customizeMessage($message);

            $target = $this->target();
            $targetLink = "<a href=\"$target\">";

            $patterns = array(
                ',%TARGET_URL%,'    => $target,
                ',%TARGET_A_OPEN% *,' => $targetLink,
                ', *%TARGET_A_CLOSE%,' => '</a>',
            );

            $this->_cookedMessage = preg_replace(array_keys($patterns), array_values($patterns), $message);
        }

        return $this->_cookedMessage;
    }

    public function performRedirection()
    {
        $target = $this->target();

        if (!$target) {
            $this->debugLog(t('No target set for this rule, even though it was activated, nothing will happen'));

            return;
        }
        $this->debugLog(t("rule activated, target is set to '%s'", $target));

        switch ($this->_redirectType) {
            case self::NO_REDIRECT :
                $this->debugLog(t("no redirection... doing nothing"));

                return;

            case self::REDIRECT_FORCED :
                if (!$target) {
                    return;
                }
                $this->debugLog(t("performing http redirection to %s", $target));

                if (self::TARGET_URL == $this->_targetType) {
                    header("Location: ".$target);
                    exit;
                } else {
                    AbstractController::redirect($target); // exits after redirect
                }

            case self::REDIRECT_SUGGEST :
                $uh = Core::make('helper/concrete/urls');
                $html = Core::make('helper/html');

                $layout = $this->notifyLayout();
                $theme  = $this->notifyTheme();
                $rID = $this->_rID;
                $view = View::getInstance();

                $view->requireAsset('javascript', 'jquery');

                // TODO (Maybe for later improvement): Must register this as assets (would be better, but this still actually works)

                $view->addFooterItem($html->javascript("3rdparty/noty/jquery.noty.js", 'gnt_maxmind'));
                $view->addFooterItem($html->javascript("3rdparty/noty/layouts/$layout.js", 'gnt_maxmind'));
                $view->addFooterItem($html->javascript("3rdparty/noty/themes/$theme.js", 'gnt_maxmind'));

                // JS notifier is now directly inlined in the page with a dedicated asset, but still added the old way
                $view->addFooterItem(new NotifierAsset($rID));

                break;
        }
    }

    public function target()
    {
        switch ($this->_targetType) {
            case self::TARGET_NONE: return;
            case self::TARGET_URL : return $this->_link;
            case self::TARGET_PAGE :

                // Prevent recursion :
                $c = Page::getCurrentPage();
                if ($c && !$c->isError() && $c->getCollectionID() == $this->_targetCID) {
                    return;
                }

                $p = Page::getByID($this->_targetCID);
                if (!$p || $p->isError()) {
                    return;
                }

                $nh = Core::make('helper/navigation');
                $link = $nh->getLinkToCollection($p);
                if ($this->_params) {
                    $sep = '&';
                    if (!strstr($link, '?')) {
                        $sep = '?';
                    }
                    $link = "$link$sep$this->_params";
                }

                return $link;
        }
    }

    protected static function _isList($ref)
    {
        if (preg_match('/[^,]+,/', $ref)) {
            return true;
        }
        return false;
    }

    protected static function _getItems($ref)
    {
        $items = preg_split('/,/', $ref);
        foreach ($items as $key => $item) {
            $items[$key] = trim($item);
        }
        return $item;
    }

    protected static function _isRange($ref)
    {
        return preg_match('/[^\-]+-[^\-]/', $ref);
    }


    protected static function _isIPv4Network($ref)
    {
        return preg_match('@\d{1,3}(\.\d{1,3}){3}/\d{1,2}@', $ref);
    }

    protected static function _splitRange($ref)
    {
        $items = preg_split('/-/', $ref);
        foreach ($items as $key => $item) {
            $items[$key] = trim($item);
        }
        return $items;
    }

    protected static function _calcIPv4Range($ref)
    {
        $r = preg_split(',/,', $ref);
        $base = $r[0];
        $mask = intval($r[1]);

        $numBase = ip2long($base);
        $ipMask = ~((1<<(32-$mask))-1);

        $firstNum = ($numBase&$ipMask) + 1;
        $first = long2ip($firstNum);
        $last = long2ip(($firstNum|~$ipMask)-1);

        return array($first, $last);
    }


    protected function _matchEqual($datumType, $dat, $ref)
    {
        $reference = strtolower($ref);
        $data = trim(strtolower($dat));

        if (self::_isList($reference)) {

            Log::debug(t("Matching a list of items in $reference")); // XXX

            foreach(self::_getItems($reference) as $operand) {
                if ($data == $operand) {
                    return true;
                }
                return false;
            }
        } 
        
        $range = null;
        if (self::_isRange($reference)) {
            $range = self::_splitRange($reference);
            $this->debugLog(t("Matching a range in %s (%s to %s)", $reference, $range[0], $range[1]));
        } else if ($datumType == self::DATUM_IP && self::_isIPv4Network($reference)) {
            $range = self::_calcIPv4Range($reference);
            $this->debugLog(t("Matching a network notation in %s (%s to %s)", $reference, $range[0], $range[1])); 
        }

        if (is_null($range)) {
            return $data == $reference;
        }

        if (self::DATUM_IP==$datumType) {
            $data = ip2long($data);
            $range[0] = ip2long($range[0]);
            $range[1] = ip2long($range[1]);
        }

        return $data <= $range[1] && $data >= $range[0];
    }

    protected function _match($dType, $type, $data, $reference)
    {
        switch ($type) {
            case self::MATCH_NOMATCH : return null;
            case self::MATCH_EQUAL   : return $this->_matchEqual($dType, $data, $reference);
            case self::MATCH_DIFFER  : return !$this->_matchEqual($dType, $data, $reference);
            case self::MATCH_REGEX :
                $pattern = str_replace(',', '\,', $reference);
                return (preg_match(",$pattern,", $data) ? true : false);
        }
    }

    public function route($ip = null, $countryCode = null, $stateCode = null, $city = null)
    {
        if (!$this->_activated) {
            $this->debugLog(t("rule is deactivated ... ignoring incoming visitor"));
            return false;
        }

        if ($this->_excludeAdmins || $this->_excludeUsers) {
            $user = new User();
            if ($user->isLoggedIn()) {

                if ($this->_excludeUsers) {
                    $this->debugLog(t('user is logged in, and rule is set to ignore logged in users ... ignoring incoming visitor'));
                    return false;
                }

                if ($user->superUser) {
                    $this->debugLog(t('user is admin, and rule is set to ignore admins... ignoring incoming visitor'));
                    return false;
                }
            }
        }

        $matchesDesc = array(
            t('country code') => array(self::DATUM_COUNTRY, $this->_countryMatchType, trim(strtolower($countryCode)), trim($this->_countryCode)),
            t('state code')   => array(self::DATUM_STATE,   $this->_stateMatchType,   trim(strtolower($stateCode)),   trim($this->_state)),
            t('city')         => array(self::DATUM_CITY,    $this->_cityMatchType,    trim(strtolower($city)),        trim($this->_cityName)),
            t('ip')           => array(self::DATUM_IP,      $this->_ipMatchType,      trim(strtolower($ip)),          trim($this->_ip)),
        );

        $matches = array();
        foreach ($matchesDesc as $name => $m) {
            list($dType, $type, $data, $ref) = $m;


            if ($this->debug()) {
                // make the assertions actually human readable (for logging)
                $assertions = array(
                    self::MATCH_NOMATCH => t('no match'),
                    self::MATCH_EQUAL   => t("visitor: '%s' is equal to or in ref: '%s'",  $data, strtolower($ref)),
                    self::MATCH_DIFFER  => t("visitor: '%s' differs from ref: '%s'", $data, strtolower($ref)),
                    self::MATCH_REGEX   => t("visitor: '%s' is matched by /%s/ regular expression", $data, $ref),
                );
                $assertion = $assertions[$type];
            }

            $match = $this->_match($dType, $type, $data, $ref);

            if ($match === true) {
                $matches[] = $match;
                $this->debugLog(t('match on %s: assertion - %s - verified', $name, $assertion));
            }
            
            if ($match === false) {
                $this->debugLog(t('mismatch on %s: assertion - %s - failed (rejecting rule)', $name, $assertion));
                return false;
            }
        }

        if (count($matches)) {
            $this->debugLog(t('some assertions were verified, and no assertion rejected, rule is activated'));
            return true;
        }

        $this->debugLog(t('no assertion successfully matched the incoming visitor, rule is skipped'));
    }

    protected $_fields = array(
        'rsID', 
        'countryCode', 'countryMatchType', 
        'stateCode', 'stateMatchType',
        'cityName', 'cityMatchType',
        'redirectType', 'targetType', 'rank',
        'targetCID', 'timeout', 'link', 'params', 'message', 'notifyLayout', 'notifyType', 'notifyTheme', 'ipMatchType', 'ip',
    );

    protected $_checkboxes = array( 'excludeUsers', 'excludeAdmins', 'activated', 'hideOnTimeout', 'notifyModal', 'useDefaultMessage' );

    protected function _getFields()
    {
        return array_merge($this->_fields, $this->_checkboxes);
    }

    public function setPropertiesFromArray(array $array)
    {
        $fields = $this->_getFields();
        if (!is_null($this->_rsID) || !$array['rsID']) {
            unset($fields['rsID']);
        }
        if (is_null($this->_rID) || !$array['rID']) {
            $fields[] = 'rID';
        }
        foreach ($fields as $key) {
            $property = "_$key";
            if (in_array($key, $this->_checkboxes)) {
                $this->$property = 0;
            }
            if (isset($array[$key])) {
                $this->$property = $array["$key"];
            }
        }

        $limits = array(
            'link' => 255, 'params' => 255, 'ip' => 255, 'countryCode' => 255, 'cityName' => 255,
        );

        foreach ($limits as $key => $sz) {
            $property = "_$key";
            if (isset($this->$property) && count($property)>$sz) {
                $this->$property = substr($this->$property, 0, $sz);
            }
        }
    }

    public function getRecord()
    {
        $record = array();
        $fields = $this->_getFields();
        $fields[] = 'rID';
        foreach ($fields as $key) {
            $property = "_$key";
            $record[$key] = $this->$property;
        }

        return $record;
    }

    protected $_error = null;
    protected function _saveError($err)
    {
        $this->_error = $err;
    }
    public function error()
    {
        return $this->_error;
    }

    public function validate()
    {
        if ($this->_targetType == self::TARGET_PAGE && !$this->_targetCID) {
            $this->_saveError(t('target page can not be empty'));

            return false;
        }

        if ($this->_targetType == self::TARGET_URL && !$this->_link) {
            $this->_saveError(t('target link can not be empty'));

            return false;
        }

        if ($this->_ipMatchType != self::MATCH_NOMATCH && !$this->_ip) {
            $this->_saveError(t('ip can not be empty'));

            return false;
        }

        return true;
    }

    public function save()
    {
        if (!$this->_rsID && $this->_ruleSet) {
            $this->_rsID = $this->_ruleSet->getRuleSetID();
        }

        $record = $this->getRecord();
        unset($record['rID']);

        $db = Database::getActiveConnection();

        if ($this->validate() !== true) {
            return false;
        }

        if (is_null($this->_rID)) {
            $db->insert('gntMaxMindRedirectRule', $record);
            $rs = $db->Execute('SELECT LAST_INSERT_ID()');
            $row = $rs->FetchRow();
            $this->_rID = $row['LAST_INSERT_ID()'];
        } else {
            $db->update('gntMaxMindRedirectRule', $record, array('rID' => $this->_rID));
        }
    }

    public function delete()
    {
        if (!$this->_rID) {
            return;
        }
        $q = "DELETE FROM gntMaxMindRedirectRule WHERE rID = $this->_rID";
        Database::getActiveConnection()->Execute($q);
        $this->_rID = null;
    }

    protected static function _getFromQuery($q, array $params = array())
    {
        $db = Database::getActiveConnection();
        $r = $db->Execute($q, $params);
        if ($r->RowCount() == 0) {
            return;
        }

        $row = $r->FetchRow();

        $rule = new RedirectRule();
        $rule->setPropertiesFromArray($row);

        return $rule;
    }

    public static function getByID($id)
    {
        if (!$id) {
            return;
        }

        return self::_getFromQuery("SELECT * FROM gntMaxMindRedirectRule WHERE rID = ?", array($id));
    }
}

